import ij.*;
import ij.gui.*;
import ij.io.*;
import ij.plugin.TextReader;
import ij.plugin.filter.ExtendedPlugInFilter;
import ij.plugin.filter.PlugInFilter;
import ij.plugin.filter.PlugInFilterRunner;
import ij.plugin.frame.Recorder;
import ij.process.*;
import ij.util.Tools;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.*;
import java.io.*;
import java.util.Arrays;

public class LOCO_ implements ExtendedPlugInFilter, DialogListener, ActionListener {
	
	Button openKernel;
	int nPasses = 1;
	int pass;
	boolean canceled;
	boolean kernelError;
	private static double[] kernel; 
	Roi roi;
	
	private ImagePlus imp;
	private int flags = DOES_16|FINAL_PROCESSING;
	private static byte Nangles = 20;
	private static int thrlevel = 0;
	private static boolean autothrFlag = true;
	private static int kernelWidth = 3;
	private static int kernelHeight = 21;
	private static double zeroAngle = 0.0;
	private static int ImageWidth, ImageHeight, ImageDimension;
	private static int RectangleX, RectangleY, RectangleW, RectangleH, RectangleSize; 
	private static String kernelText;
	private static boolean openkernelFlag = false;
	private static boolean orientStackFlag = true;
	private static boolean fullIstackFlag = true;
	private static boolean intStackFlag = true;
	private static boolean origStackFlag = true;
	private static double[] xref_kernel,yref_kernel;
	private static int i_roi = 0;
	private static int[] xinRoi ;
	private static int[] yinRoi ;
	private static int background = 0; 
	private static long startTime=0;
	
	//developer flags
	private static boolean maskFlag = false;
	
	GenericDialog gd;
	PlugInFilterRunner pfr;
	
	private static ImageStack maxIstack,fullIstack,maxIAnglestack,origIstack;
	private boolean ctrlVisual = true;
	
	public int setup(String arg, ImagePlus imp) {
		this.imp = imp; 
		if (imp==null){
			IJ.noImage(); 
			return DONE;
		}
		if (arg.equals("final")) {					
			finalprocessing();
			return DONE;
			
		} else {
			initialprocessing(imp);
			return flags;
		}
	}
	
	public void initialprocessing(ImagePlus imp){

		imp.startTiming();
		
		origIstack = imp.getStack();
		ImageWidth = imp.getWidth();
		ImageHeight = imp.getHeight();
		ImageDimension = ImageHeight*ImageWidth;
			
		roi=imp.getRoi();
		if((roi==null)||(roi.getType()>2)) {
			System.out.println("Please, select simple ROI. Now switch to whole image ROI");
			imp.setRoi(0,0, ImageWidth,ImageHeight);
			roi=imp.getRoi();
		}
		startTime=imp.getStartTime();
		
		handleRoi(roi);		
		createStacks();
	
		IJ.resetEscape();
	}
	
	public void handleRoi(Roi roi){					
		
		boolean rectRoi=false;
		if(roi.getType()==0) rectRoi=true;
		
		Rectangle r = roi.getBounds(); 
		RectangleW = r.width;
		RectangleH = r.height;
		RectangleX = r.x;
		RectangleY = r.y;
		RectangleSize = RectangleW*RectangleH;

		xinRoi = new int [RectangleSize];
		yinRoi = new int [RectangleSize];
		
		i_roi = 0;
		byte[] tmpMaskRoi = new byte[RectangleSize];
		
		byte wh = (byte) 255; 
		if(!rectRoi) {									
			tmpMaskRoi = (byte[]) roi.getMask().getPixels();
			
			int i_xy = 0;
			for (int y=0;y<RectangleH;y++){					    	
				for (int x=0;x<RectangleW;x++){	
					if (tmpMaskRoi[i_xy] == wh) {
						xinRoi[i_roi] = x;
						yinRoi[i_roi] = y;
						i_roi++;
					}
					i_xy++;
				}
			}
		} else {
			for (int y=0;y<RectangleH;y++){					    	
				for (int x=0;x<RectangleW;x++){
					xinRoi[i_roi] = x;
					yinRoi[i_roi] = y;
					i_roi++;
				}
			}
		}
		
		if(maskFlag) new ImagePlus("Mask", new ByteProcessor(RectangleW,RectangleH,tmpMaskRoi,null)).show();
	}	
	
	public void createStacks(){
		maxIstack = new ImageStack(RectangleW,RectangleH);
		if (fullIstackFlag) fullIstack = new ImageStack(RectangleW,RectangleH);
		maxIAnglestack = new ImageStack(RectangleW,RectangleH);
		
	}

	public int showDialog(ImagePlus imp, String command, PlugInFilterRunner pfr) {
		this.pfr = pfr;
		gd = new GenericDialog("Rotating Kernel",IJ.getInstance());
		gd.addMessage("Width:    Height:");
		gd.addTextAreas(String.valueOf(kernelWidth), String.valueOf(kernelHeight), 1, 4);
		gd.addMessage("--------------");
		gd.addPanel(makeButtonPanel(gd));
		gd.addNumericField("Number of angles: ", Nangles, 0);
		gd.addMessage("--------------");
		gd.addMessage("Show");
		gd.addCheckbox("Max orientation stack", false);
		gd.addCheckbox("Full orientation stack", false);
		gd.addCheckbox("Average intensity stack", false);
		gd.addMessage("--------------\n");
		gd.addCheckbox("Threshold original image", true);
		gd.addCheckbox("Auto threshold", true);	
		gd.addNumericField("User threshold %", 0,0);
		gd.addNumericField("Reference angle", zeroAngle, 0);
		gd.addDialogListener(this);
		gd.showDialog();
		if (gd.wasCanceled()) return DONE;
		return IJ.setupDialog(imp, flags); 
	}
	
	Panel makeButtonPanel(GenericDialog gd) {
		Panel buttons = new Panel();
    	buttons.setLayout(new FlowLayout(FlowLayout.CENTER, 5, 0));
		openKernel = new Button("Load user kernel...");
		openKernel.addActionListener(this);
		buttons.add(openKernel);
		return buttons;
	}
	
	public boolean dialogItemChanged(GenericDialog gd, AWTEvent e) {
		try {
			kernelWidth = Integer.parseInt(gd.getTextArea1().getText().trim());
			kernelHeight = Integer.parseInt(gd.getTextArea2().getText().trim());
		} catch (NumberFormatException nfe) {
			System.out.println("NumberFormatException: " + nfe.getMessage()); 
		}
		Nangles = (byte)gd.getNextNumber();
		orientStackFlag = gd.getNextBoolean();
		fullIstackFlag = gd.getNextBoolean();
		intStackFlag = gd.getNextBoolean();
		origStackFlag = gd.getNextBoolean();
		autothrFlag = gd.getNextBoolean();
		int thrlevel_tmp = (int)gd.getNextNumber();
		if(!autothrFlag){
			thrlevel = thrlevel_tmp;
		}
		zeroAngle = gd.getNextNumber();
		if (gd.invalidNumber()) { 
			if (gd.wasOKed()) IJ.error("Some number is invalid.");
			return false;
		}
		if(openkernelFlag){ 
			kernelError = !decodeKernel(kernelText);
		} else {
			int i = 0;
			int kernelNorm = kernelWidth*kernelHeight;
			kernel = new double[kernelNorm]; 
			for (int y=0; y<kernelHeight; y++) {
				for (int x=0; x<kernelWidth; x++){
					kernel[i] = 1.0/kernelNorm;
					i++;
				}
			}
			kernelError=false;
		}
		if (!kernelError) {
			IJ.showStatus("Rotating convolution: "+kernelWidth+"x"+kernelHeight+" kernel");
			return true;
		}
		return true;

	}
	
	
	public void actionPerformed(ActionEvent e) {
		Object source = e.getSource();
		Recorder.disablePathRecording();
		if (source==openKernel) open();
		openkernelFlag = true; 
	}
	
	void open() {
		OpenDialog od = new OpenDialog("Open Kernel...", "");
		String directory = od.getDirectory();
		String name = od.getFileName();
		if (name==null)
			return;
		String path = directory + name;
		TextReader tr = new TextReader();
		ImageProcessor ip = tr.open(path);
		if (ip==null)
			return;
		int width = ip.getWidth();
		int height = ip.getHeight();
		if ((width&1)!=1 || (height&1)!=1) {
			IJ.error("Rotating Convolver", "Kernel must be have odd width and height");
			return;
		}
		StringBuffer sb = new StringBuffer();
		boolean integers = true;
		for (int y=0; y<height; y++) {
			for (int x=0; x<width; x++) {
				double v = ip.getPixelValue(x, y); 
				if ((int)v!=v)
					integers = false;
			}
		}
		for (int y=0; y<height; y++) { 
			for (int x=0; x<width; x++) {
				if (x!=0) sb.append(" ");
				double v = ip.getPixelValue(x, y);
				if (integers)
					sb.append(IJ.d2s(v,0));
				else
					sb.append(""+v);
			}
			if (y!=height-1)
				sb.append("\n");
		}
		kernelText = new String(sb);
		gd.getTextArea1().setText(String.valueOf(width));
		gd.getTextArea2().setText(String.valueOf(height));
		gd.getComponent(0).setVisible(false); 
		gd.setTitle("Rotating User Kernel");
	}
	
	boolean decodeKernel(String text) {
		String[] rows = Tools.split(text, "\n");
		kernelHeight = rows.length;
		if (kernelHeight==0) return false;
		String[] values = Tools.split(rows[0]);
		kernelWidth = values.length;
		if (kernelWidth==0) return false;
		int kernelArea = kernelWidth*kernelHeight;
		kernel = new double[kernelArea];
		boolean done = gd.wasOKed();
		int i = 0;
		for (int y=0; y<kernelHeight; y++) {
			values = Tools.split(rows[y]);	
			if (values.length!=kernelWidth) {
				String err = "Row "+(y+1)+" is not the same length as the first row";
				if (done)
					IJ.error("Rotating Convolver", err);
				else
					IJ.showStatus(err);
				return false;
			}
			for (int x=0; x<kernelWidth; x++)
				kernel[i++] = (float)Tools.parseDouble(values[x], 0.0);
		}
		if ((kernelWidth&1)!=1 || (kernelHeight&1)!=1) {
			String err = "Kernel must have odd width and height. This one is "+kernelWidth+"x"+kernelHeight+".";
			if (done)
				IJ.error("Rotating Convolver", err);
			else
				IJ.showStatus(err);
			return false;
		}
		double kernelNorm = 0;
		for (int j=0;j<kernelArea;j++){
			kernelNorm = kernelNorm + kernel[j];
		}
		if (kernelNorm>1e-06) {
			for (int j=0;j<kernelArea;j++){
				kernel[j] = kernel[j]/kernelNorm;
			}
		}
		return true;
    }
	
	public void setNPasses(int nPasses) {
		this.nPasses = nPasses;
		pass = 0;
	}
	
	void showProgress(double percent) {
		percent = (double)(pass-1)/nPasses + percent/nPasses;
		IJ.showProgress(percent);
	}
	
	public void run(ImageProcessor ip) {
		
		if (pass==0) {
			xyref_kernel();
		}
		if (canceled) return;
		
		kernelConvolution(ip,pass);
		pass++;
	}
	
	void xyref_kernel (){
		int Nall = Nangles*kernelHeight*kernelWidth;
		xref_kernel = new double [Nall];
		Arrays.fill(xref_kernel, (double)RectangleX);
		yref_kernel = new double [Nall];
		Arrays.fill(yref_kernel, (double)RectangleY);
		
		int fh2 = (kernelHeight - 1)/2;
		int fw2 = (kernelWidth - 1)/2;
		
		int i = 0;
		int shiftAngle = Nangles / 2;
		int shiftedAngle = shiftAngle;
		
		for (int angle=Nangles;angle>0;angle--){	
			
			if (shiftAngle <= Nangles){
				shiftedAngle = shiftAngle;
			}
			else 
			{
				shiftedAngle = shiftAngle - Nangles;
			}	
			
			double a = (Math.PI*shiftedAngle)/Nangles;
			double ca = Math.cos(a); //or here
			double sa = Math.sin(a);

			for (int mx=-fw2;mx<=fw2;mx++){
				double s1 = mx*ca;
				double s2 = mx*sa;
				   	
				for (int my=-fh2;my<=fh2;my++){
					xref_kernel[i] += (s1 - my*sa);
					yref_kernel[i] += (s2 + my*ca);
					i++;
				}
			}
			shiftAngle++;
		}
	}
	
	public void kernelConvolution(ImageProcessor ip, int pass){
		
		short[] maxOrientIntensity = new short[RectangleSize];
		byte[] AnglemaxOrientIntensity = new byte[RectangleSize];
		short[] OrientIntensity = new short[RectangleSize*Nangles];
		
		int fh2 = (kernelHeight-1)/2;
		int fw2 = (kernelWidth-1)/2;
		
		short[] l_pixels = new short[ImageDimension];
		l_pixels = (short[]) ip.getPixels();

		for (int i=0;i<i_roi;i++){					    	
			int x = xinRoi[i]; 
			int y = yinRoi[i]; 
			int ir = y*RectangleW + x;
			int im = ir*Nangles;
			int k = 0;
			double maxIav = 0.0;
			byte maxAngle = 0;
			
			for (byte angle = Nangles;angle>0;angle--){	
				double Iav = 0;
				int ik = 0;
				
				for (int mx=-fw2;mx<=fw2;mx++){
					for (int my=-fh2;my<=fh2;my++){		
						double xa = x + xref_kernel[k];
						double ya = y + yref_kernel[k];
						
						int xbase = (int)xa;
						int ybase = (int)ya;
						double xFraction = xa - xbase;
						double yFraction = ya - ybase;	
						
						int iuL = ybase*ImageWidth + xbase;						
						int iuR = iuL+1;
						int ilL = iuL+ImageWidth;
						int ilR = iuR+ImageWidth;
						
						int lowerLeft = ((ilL<ImageDimension) && (ilL>0)) ? l_pixels[ilL]&0xffff : background;
						int lowerRight = ((ilR<ImageDimension) && (ilR>0)) ? l_pixels[ilR]&0xffff : background;
						int upperRight = ((iuR<ImageDimension) && (iuR>0)) ? l_pixels[iuR]&0xffff : background;
						int upperLeft = ((iuL<ImageDimension) && (iuL>0)) ? l_pixels[iuL]&0xffff : background;
						
						double upperAverage = upperLeft + xFraction * (upperRight - upperLeft);
						double lowerAverage = lowerLeft + xFraction * (lowerRight - lowerLeft);
					
						Iav += (lowerAverage + yFraction * (upperAverage - lowerAverage))*kernel[ik];
						ik++;
						k++;
					}
				}
				if (Iav>maxIav) {maxIav = Iav; maxAngle = angle;}
				
				OrientIntensity[im] = (short)Iav;				
				im++;
			}
			
			maxOrientIntensity[ir] = (short)maxIav;
			AnglemaxOrientIntensity[ir] = maxAngle;
		}
		String slname = Integer.toString(pass);
		
		maxIstack.addSlice(slname, maxOrientIntensity);
		maxIAnglestack.addSlice(slname, AnglemaxOrientIntensity);
		
		if (fullIstackFlag){
			for (byte angle = Nangles;angle>0;angle--){
				short[] tIntensity = new short[RectangleSize];
				for (int px = 0;px<RectangleSize;px++){
					tIntensity[px]=OrientIntensity[px*Nangles+angle-1];
				}
				tIntensity[0]=angle;
				fullIstack.addSlice(slname, tIntensity); 			
			}
		}
	}
	
	public void finalprocessing(){
		
		if (intStackFlag) new ImagePlus("maxI",maxIstack).show();
		if (orientStackFlag) new ImagePlus("maxIAngle",maxIAnglestack).show();
		if (fullIstackFlag){
			new ImagePlus("fullIAngle",fullIstack).show();
			int Nslices=maxIstack.getSize();
			if(Nslices>1) {
				IJ.selectWindow("fullIAngle");
				String myCommand="Stack to Hyperstack...";
				String myParams="order=xyczt(default) channels=1 slices="+Integer.toString(Nangles)+
					" frames="+Integer.toString(Nslices) +" display=Grayscale";
				IJ.run(myCommand,myParams);
			}
		}
		if(ctrlVisual) new ImagePlus("name",ResultStackAnalysis()).show();	
		
		System.out.println("Done in "+(System.currentTimeMillis()-startTime)*0.001+" seconds\n");
	}		

	public ImageStack ResultStackAnalysis(){
		ImageProcessor trip;
		ByteProcessor bip;
		ImageStack tstk = new ImageStack(RectangleW,RectangleH);
		byte[] TreshBaseImagePixels = new byte [RectangleSize];
		byte[] AnglePixels = new byte [RectangleSize];
		byte nol = 0;				
		
		ImageStack tmpstk = (origStackFlag)? origIstack : maxIstack;
		int nslices = maxIstack.getSize();
		
		for (int slice=1;slice<=nslices;slice++){
			trip = tmpstk.getProcessor(slice).duplicate();

			if(autothrFlag) {
				trip.autoThreshold();
			} else {
				trip.threshold(thrlevel);	
			}														
			bip = (ByteProcessor) trip.convertToByte(false);
			
			AnglePixels = (byte[]) maxIAnglestack.getProcessor(slice).getPixelsCopy();
			TreshBaseImagePixels = (byte[])bip.getPixels();
			
			for (int k=0;k<RectangleSize;k++){
				AnglePixels[k]=(TreshBaseImagePixels[k]==nol)? 0 : AnglePixels[k];
			}
			
			tstk.addSlice(Integer.toString(slice), AnglePixels);			
		}
		return tstk;
	}	
}
